import unittest
from src.pyphetools.creation import Thresholder



class TestThresholder(unittest.TestCase):

    def test_alk_phos(self):
        # 30-120
        apThresholder = Thresholder.alkaline_phophatase_blood()
        hpterm = apThresholder.map_value(12)
        # Low alkaline phosphatase (HP:0003282)
        self.assertEqual("Low alkaline phosphatase", hpterm.label)
        self.assertEqual("HP:0003282", hpterm.id)
        self.assertTrue(hpterm.observed)
        # Elevated circulating alkaline phosphatase concentration (HP:0003155)
        hpterm = apThresholder.map_value(500)
        self.assertEqual("Elevated circulating alkaline phosphatase concentration", hpterm.label)
        self.assertEqual("HP:0003155", hpterm.id)
        self.assertTrue(hpterm.observed)
        # Abnormality of alkaline phosphatase level (HP:0004379)
        hpterm = apThresholder.map_value(60) # normal value
        self.assertEqual("Abnormality of alkaline phosphatase level", hpterm.label)
        self.assertEqual("HP:0004379", hpterm.id)
        self.assertFalse(hpterm.observed)
        print(hpterm)

    def test_potassium(self):
        kThresholder = Thresholder.potassium_blood()
        hpterm = kThresholder.map_value(2)
        self.assertEqual("Hypokalemia", hpterm.label)
        self.assertEqual("HP:0002900", hpterm.id)

    def test_NTproBNP_blood(self):
        thresholder = Thresholder.NTproBNP_blood()
        hpterm = thresholder.map_value(200)
        self.assertEqual("Increased circulating NT-proBNP concentration", hpterm.label)
        self.assertEqual("HP:0031185", hpterm.id)



import unittest
from src.pyphetools.creation import ConstantColumnMapper



class TestConstantMapper(unittest.TestCase):

    def test_observed_constant(self):
        hp_id =  "HP:0031956"
        hp_label = "Elevated circulating aspartate aminotransferase concentration"
        mapper = ConstantColumnMapper(column_name="placeholder",hpo_id=hp_id, hpo_label=hp_label)
        hp_term_list = mapper.map_cell("n/a")
        self.assertEqual(1, len(hp_term_list))
        result = hp_term_list[0]
        self.assertEqual(hp_id, result.id)
        self.assertEqual(hp_label, result.label)
        self.assertTrue(result.observed)

    def test_excluded_constant(self):
        hp_id =  "HP:0031956"
        hp_label = "Elevated circulating aspartate aminotransferase concentration"
        mapper = ConstantColumnMapper(column_name="placeholder", hpo_id=hp_id, hpo_label=hp_label, excluded=True)
        hp_term_list = mapper.map_cell("n/a")
        self.assertEqual(1, len(hp_term_list))
        result = hp_term_list[0]
        self.assertEqual(hp_id, result.id)
        self.assertEqual(hp_label, result.label)
        self.assertFalse(result.observed)
        self.assertTrue(result.excluded)



import re
from collections import defaultdict
from enum import Enum

import pandas as pd

from .age_isoformater import AgeIsoFormater
from .constants import Constants

AgeEncodingType = Enum('AgeEncodingType', ['YEAR', 'ISO8601', 'YEAR_AND_MONTH', 'MONTH', 'CUSTOM', 'NOT_PROVIDED'])
ISO8601_REGEX = r"^P(\d+Y)?(\d+M)?(\d+D)?"
# e.g., 14 y 8 m or 8 y
YEAR_AND_MONTH_REGEX = r"(\d+)\s*[Yy]\s*(\d+)\s*[Mm]"
YEAR_REGEX = r"(\d+)\s*[Yy]"
MONTH_REGEX = r"(\d+)\s*[Mm]"


class AgeColumnMapper:
    """
    Map a column that contains information about the age of individuals.

    Tables with information about genotype phenotype correlations typically
    contain a column with information about age. The columns often have formats
    such as 34 (integer with number of years or months) or 3Y2M (for three years
    and two months). This mapper ingests data from such columns and transforms the
    contents into ISO 8601 strings (e.g., P4Y2M1D for 4 years, 2 months, and 1 day).

    """

    def __init__(self, ageEncodingType, column_name, string_to_iso_d=None) -> None:
        """
        :param ageEncodingType: Formatting convention used to represent the age
        :type ageEncodingType: one of Year (e.g. 42), ISO 8601 (e.g. P42Y2M), year/month (e.g. 42y2m)
        :param column_name: Name of the Age column in the original table
        :type column_name: str
        :param string_to_iso_d: dictionary from free text (input table) to ISO8601 strings
        :type string_to_iso_d: Dict[str,str], optional
        """

        if string_to_iso_d is None:
            string_to_iso_d = {}
        self._age_econding = ageEncodingType
        if column_name is None:
            raise ValueError("Must provide non-null column_name argument")
        self._column_name = column_name
        self._string_to_iso_d = string_to_iso_d
        self._erroneous_input_counter = defaultdict(int)

    def map_cell(self, cell_contents) -> str:

        """
        Map a single cell of the table

        :param cell_contents: The text contained in a single cell of the table
        :type cell_contents: can be a string or numerical type
        """

        contents = str(cell_contents)
        contents = contents.strip()
        if self._age_econding == AgeEncodingType.YEAR:
            isostring = self.get_iso8601_from_int_or_float_year(contents)
            if isostring is None:
                print(f"Could not parse {contents} as integer (year): {cell_contents}")
            return isostring
        elif self._age_econding == AgeEncodingType.YEAR_AND_MONTH:
            try:
                match = re.search(YEAR_AND_MONTH_REGEX, contents)
                if match:
                    years = int(match.group(1))
                    months = int(match.group(2))
                    return AgeIsoFormater.to_string(y=years, m=months)
                match = re.search(YEAR_REGEX, contents)
                if match:
                    years = int(match.group(1))
                    return AgeIsoFormater.to_string(y=years)
                match = re.search(MONTH_REGEX, contents)
                if match:
                    months = int(match.group(1))
                    return AgeIsoFormater.to_string(m=months)
            except ValueError as verr:
                print(f"Could not parse {contents} as year/month: {verr}")
                return Constants.NOT_PROVIDED
        elif self._age_econding == AgeEncodingType.MONTH:
            # assume month encoded by integer or float.
            return AgeIsoFormater.from_numerical_month(contents)
        elif self._age_econding == AgeEncodingType.ISO8601:
            match = re.search(ISO8601_REGEX, contents)
            if match:
                return contents
            else:
                self._erroneous_input_counter[contents] += 1
                return Constants.NOT_PROVIDED
        elif self._age_econding == AgeEncodingType.CUSTOM:
            return self._string_to_iso_d.get(cell_contents, Constants.NOT_PROVIDED)
        elif self._age_econding == AgeEncodingType.NOT_PROVIDED:
            return Constants.NOT_PROVIDED

    def get_iso8601_from_int_or_float_year(self, age) -> str:

        """
        Extract an iso8601 string for age recorded as a year (either an int such as 4 or a float such as 4.25 for P4Y3M)
        :param age: an int representing years or a float such as 2.5 for two and a half years
        :return: an ISO 8601 string such as P2Y6M
        """

        if isinstance(age, int):
            return f"P{age}Y"
        elif isinstance(age, float):
            age = str(age)
        elif not isinstance(age, str):
            raise ValueError(f"Malformed agestring {age}, type={type(age)}")
        int_or_float = r"(\d+)(\.\d+)?"
        p = re.compile(int_or_float)
        results = p.search(age).groups()
        if len(results) != 2:
            return Constants.NOT_PROVIDED
        if results[0] is None:
            return Constants.NOT_PROVIDED
        y = int(results[0])
        if results[1] is None:
            return f"P{y}Y"
        else:
            m = float(results[1])  # something like .25
            months = round(12 * m)
            return f"P{y}Y{months}M"

    def preview_column(self, df:pd.DataFrame) -> pd.DataFrame:
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        preview_d = defaultdict(list)
        column = df[self.get_column_name()]
        for _, value in column.items():
            preview_d[value] = self.map_cell(str(value))
        dlist = []
        for k, v in preview_d.items():
            if v is None:
                dlist.append({"original column contents": k, "age": "n/a"})
            else:
                dlist.append({"original column contents": k, "age": v})
        return pd.DataFrame(dlist)

    def get_column_name(self) -> str:
        return self._column_name

    def has_error(self) -> bool:
        return len(self._erroneous_input_counter) > 0

    def error_summary(self):
        items = []
        for k, v in self._erroneous_input_counter.items():
            items.append(f"{k} (n={v})")
        return f"Could not parse the following as ISO8601 ages: {', '.join(items)}"


    @staticmethod
    def not_provided():
        """Create an object for cases where Age is not provided.
        """
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.NOT_PROVIDED, column_name=Constants.NOT_PROVIDED)

    @staticmethod
    def by_year(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.YEAR, column_name=column_name)

    @staticmethod
    def by_year_and_month(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.YEAR_AND_MONTH, column_name=column_name)

    @staticmethod
    def by_month(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.MONTH, column_name=column_name)

    @staticmethod
    def iso8601(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.ISO8601, column_name=column_name)

    @staticmethod
    def custom_dictionary(column_name, string_to_iso_d):
        """
        Create an AgeColumnMapper for free text input data such as Fetus, 1.5, birth, 51 days
        :param column_name: name of the age column in the input table
        :type column_name: str
        :param string_to_iso_d: dictionary with free text to ISO 8601
        :type string_to_iso_d: Dict[str,str)
        """
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.CUSTOM,
                            column_name=column_name,
                            string_to_iso_d=string_to_iso_d)



import os
import phenopackets
from google.protobuf.json_format import Parse
import json
from collections import defaultdict
from ..creation.constants import Constants
from ..creation.hp_term import HpTerm
from .simple_variant import SimpleVariant

class SimplePatient:
    """
    This class flattens all observed terms into a set and also recorded variants, sex, identifier, and age
    The class purposefully disregards information about the time course in order to be able to count the
    frequencies of HPO terms in groups. The use case of the class is to facilitate visualization of
    a collection of phenopackets from files or that have just been ingested using pyphetools. Each simple
    patient is essentially a wrapper around one phenopacket.

    :param ga4gh_phenopacket: A Phenopacket object representing one individual
    :type ga4gh_phenopacket: phenopackets.schema.v2.phenopackets_pb2.Phenopacket
    """

    def __init__(self, ga4gh_phenopacket) -> None:
        if str(type(ga4gh_phenopacket)) != "<class 'phenopackets.schema.v2.phenopackets_pb2.Phenopacket'>":
            raise ValueError(f"phenopacket argument must be GA4GH Phenopacket Schema Phenopacket but was {type(ga4gh_phenopacket)}")
        else:
            ppack = ga4gh_phenopacket
        observed_hpo_terms = defaultdict(HpTerm)
        excluded_hpo_terms = defaultdict(HpTerm)
        self._by_age_dictionary = defaultdict(list)
        self._phenopacket_id = ppack.id
        if ppack.subject is None:
            raise ValueError("Phenopackets must have a subject message to be used with this package")
        subj = ppack.subject
        if subj.id is None:
            raise ValueError("Phenopacket subjects must have an id field to be used with this package")
        else:
            self._subject_id = subj.id
        if subj.time_at_last_encounter is None or subj.time_at_last_encounter.age.iso8601duration is None:
            print("Warning: No age found for phenopacket")
            self._time_at_last_encounter = None
        else:
            self._time_at_last_encounter = subj.time_at_last_encounter.age.iso8601duration
        if ppack.subject.sex == phenopackets.MALE:
            self._sex = "MALE"
        elif ppack.subject.sex == phenopackets.FEMALE:
            self._sex = "FEMALE"
        elif ppack.subject.sex == phenopackets.OTHER_SEX:
            self._sex = "OTHER"
        else:
            self._sex = "UNKNOWN"
        for pf in ppack.phenotypic_features:
            hpterm = HpTerm(hpo_id=pf.type.id, label=pf.type.label, observed=not pf.excluded)
            if pf.excluded:
                excluded_hpo_terms[pf.type.id] = hpterm
            else:
                observed_hpo_terms[pf.type.id] = hpterm
            if pf.onset is not None and pf.onset.age is not None and pf.onset.age.iso8601duration:
                term_onset = pf.onset.age.iso8601duration
            else:
                term_onset = Constants.NOT_PROVIDED
            self._by_age_dictionary[term_onset].append(hpterm)
        for k, v in observed_hpo_terms.items():
            if k in excluded_hpo_terms:
                excluded_hpo_terms.pop(k) # remove observed terms that may have been excluded at other occasion
        self._observed = observed_hpo_terms
        self._excluded = excluded_hpo_terms
        # Add information about variants
        self._variant_list = []
        self._disease = None
        if ppack.interpretations is not None and len(ppack.interpretations) > 0:
            interprets = ppack.interpretations
            for interpretation in interprets:
                if str(type(interpretation)) != "<class 'phenopackets.schema.v2.core.interpretation_pb2.Interpretation'>":
                    raise ValueError(f"interpretation argument must be GA4GH Phenopacket Interpretation but was {type(interpretation)}")
                diagnosis = interpretation.diagnosis
                if diagnosis is not None:
                    if diagnosis.disease is not None:
                        disease_id = diagnosis.disease.id
                        disease_label = diagnosis.disease.label
                        self._disease = f"{disease_label} ({disease_id})"
                ginterpretations = diagnosis.genomic_interpretations
                for gint in ginterpretations:
                    variant = SimpleVariant(ginterpretation=gint)
                    self._variant_list.append(variant)
        # Get PMID, if available, from the MetaData
        mdata = ppack.meta_data
        self._pmid = None
        if len(mdata.external_references) == 1:
            eref = mdata.external_references[0]
            self._pmid = eref.id

    @staticmethod
    def from_file(phenopacket_file):
        """
        Return a SimplePatient object that corresponds to a phenopacket (JSON) file
        :param phenopacket_file: A phenopacket file (JSON format)
        :type phenopacket_file: string representing a path to a file
        """
        if not os.path.isfile(phenopacket_file):
            raise FileNotFoundError(f"Could not find phenopacket file at '{phenopacket_file}'")
        with open(phenopacket_file) as f:
            data = f.read()
            jsondata = json.loads(data)
            ppack = Parse(json.dumps(jsondata), phenopackets.Phenopacket())
            return SimplePatient(ga4gh_phenopacket=ppack)


    @staticmethod
    def from_individual(individual, metadata):
        """
        Return a SimplePatient object that corresponds to a pyphetools Individual object
        :param individual: Am Individual object
        :type individual: Individual
        :param metadata: A GA4GH Phenopacket Schema MetaData object
        :type metadata: MetaData
        """
        ppack = individual.to_ga4gh_phenopacket(metadata)
        return SimplePatient(ga4gh_phenopacket=ppack)


    def get_phenopacket_id(self):
        return self._phenopacket_id

    def get_subject_id(self):
        return self._subject_id

    def get_sex(self):
        return self._sex

    def get_age(self):
        return self._time_at_last_encounter

    def get_disease(self):
        if self._disease is None:
            return "n/a"
        else:
            return self._disease

    def get_observed_hpo_d(self):
        """
        returns map of observed phenotypic features with key (string) HP id, value, HpTerm from creation submodule
        """
        return self._observed

    def get_excluded_hpo_d(self):
        """
        :return: map of excluded phenotypic features with key (string) HP id, value, HpTerm from creation submodule
        """
        return self._excluded

    def get_total_hpo_count(self):
        """
        :return: total count of HPO terms (observed and excluded)
        :rtype: int
        """
        return len(self._observed) + len(self._excluded)

    def get_variant_list(self):
        return self._variant_list

    def has_pmid(self):
        return self._pmid is not None

    def get_pmid(self):
        return self._pmid

    def contains_observed_term_id(self, hpo_term_id):
        return hpo_term_id in self._observed

    def get_term_by_age_dict(self):
        return self._by_age_dictionary




NA_STRING = "n/a"


class SimpleVariant:
    """Representation of a variant to facilitate display

    This class flattens information about variants from an
    GenomicInterpretation obtained from a phenopacket

    :param ginterpretation: A GenomicInterpretation object representing one variant and its interpretation
    :type ginterpretation: phenopackets.schema.v2.phenopackets_pb2.GenomicInterpretation
    """
    def __init__(self, ginterpretation) -> None:
        """
        Constructor
        """
        if str(type(ginterpretation)) != "<class 'phenopackets.schema.v2.core.interpretation_pb2.GenomicInterpretation'>":
            raise ValueError(f"interpretation argument must be GA4GH Phenopacket GenomicInterpretation but was {type(ginterpretation)}")
        self._status = str(ginterpretation.interpretation_status)
        vinterpretation = ginterpretation.variant_interpretation
        if vinterpretation.variation_descriptor is  None:
            raise ValueError("variation_descriptor was None")
        vdescript = vinterpretation.variation_descriptor
        if vdescript.gene_context is not None:
            self._gene_id = vdescript.gene_context.value_id
            self._gene_symbol = vdescript.gene_context.symbol
        else:
            self._gene_id = NA_STRING
            self._gene_symbol = NA_STRING
        self._hgvs = NA_STRING
        if vdescript.expressions is not None and len(vdescript.expressions) > 0:
            for exprsn in vdescript.expressions:
                if exprsn.syntax == "hgvs.c":
                    self._hgvs = exprsn.value
        if vdescript.vcf_record is not None and len(vdescript.vcf_record.chrom) > 0:
            vcf = vdescript.vcf_record
            self._genome_ass = vcf.genome_assembly
            self._chrom = vcf.chrom
            self._pos = int(vcf.pos)
            self._ref = vcf.ref
            self._alt = vcf.alt
        else:
            self._genome_ass = NA_STRING
            self._chrom = NA_STRING
            self._pos = -1
            self._ref = NA_STRING
            self._alt = NA_STRING
        if vdescript.allelic_state is not None:
            allelic = vdescript.allelic_state
            self._genotype_id = allelic.id
            self._genotype_label = allelic.label
        else:
            self._genotype_id = NA_STRING
            self._genotype_label = NA_STRING
        if vdescript.structural_type is not None:
            so_id = vdescript.structural_type.id
            so_label = vdescript.structural_type.label
            self._structural = f"{so_label} ({so_id})"
        else:
            self._structural = None
        self._description_label = vdescript.label

    @property
    def status(self):
        """
        :returns: interpretation status from GA4GH GenomicInterpretation
        """
        return self._status

    @property
    def gene_id(self):
        """
        :returns: the gene identifier, e.g., from HGNC such as HGNC:3603.
        :rtype: str
        """
        return self._gene_id

    @property
    def gene_symbol(self):
        """
        :returns: the gene synbol, e.g., FBN1
        :rtype: str
        """
        return self._gene_symbol

    @property
    def hgvs(self):
        """
        :returns: Human Genome Variation Society (HGVS) variant nomenclature for this variant
        :rtype: str
        """
        return self._hgvs

    @property
    def genome_assembly(self):
        """
        :returns: genome build used for reporting this variant, e.g. hg38
        :rtype: str
        """
        return self._genome_ass

    @property
    def chrom(self):
        """
        :returns: chromosome on which the variant is located
        :rtype: str
        """
        return self._chrom

    @property
    def position(self):
        """
        :returns: position of the variant on the chromosome on which it is located
        :rtype: int
        """
        return self._pos

    @property
    def ref(self):
        """
        :returns: reference sequence for variant
        :rtype: str
        """
        return self._ref

    @property
    def alt(self):
        """
        :returns: alternate sequence for variant
        :rtype: str
        """
        return self._alt

    @property
    def genotype_id(self):
        """
        :returns: the Sequence Ontology id for the genotype of this variant, optional
        :rtype: str
        """
        return self._genotype_id

    @property
    def genotype_label(self):
        """
        :returns: the Sequence Ontology label for the genotype of this variant, optional
        :rtype: str
        """
        return self._genotype_label

    def has_vcf(self):
        """
        :returns: True if there is VCF data for this variant, otherwise False
        :rtype: bool
        """
        return self._chrom != NA_STRING

    def has_genotype(self):
        """
        :returns: True if there is a genotype for this variant, otherwise False
        :rtype: bool
        """
        return self._genotype_id != NA_STRING


    def get_display(self):
        """
        :returns: a string representing a human-readable representation of the variant
        :rtype: str
        """
        if self.has_genotype():
            genotype = "(" + self._genotype_label + ")"
        else:
            genotype = ""
        if self._hgvs != NA_STRING:
            return self._hgvs + " " + genotype
        if self.has_vcf():
            vcf_str = f"{self.chrom}:{self.position}{self.ref}>{self.alt} {genotype}"
            return vcf_str
        elif self._structural is not None:
            # try to return information about a structural variant.
            var_str = f"{self._description_label}: {self._structural}"
            return var_str
        else:
            return "n/a"



import abc
from typing import Dict, List
from pyphetools.creation.citation import Citation
from pyphetools.creation.constants import Constants
from pyphetools.creation.disease import Disease
from pyphetools.creation.hpo_cr import HpoConceptRecognizer
from pyphetools.creation.metadata import MetaData
from pyphetools.creation.hp_term import HpTerm
from pyphetools.creation.individual import Individual
import os
import re
import pandas as pd
from google.protobuf.json_format import MessageToJson
import phenopackets as PPKt


# The following constants are identical with the constants used in the  Excel template
AGE_OF_ONSET_FIELDNAME = "age_of_onset"



class CellEncoder(metaclass=abc.ABCMeta):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    @abc.abstractmethod
    def encode(self, cell_contents):
        pass

    @abc.abstractmethod
    def is_data(self):
        pass

    @abc.abstractmethod
    def is_hpo(self):
        pass

class DataEncoder(CellEncoder):
    """Convenience class to represent the header (two lines) of columns that represent fixed data in the template

    :param h1: contents of the first header line
    :type h1: str
    :param h2: contents of the second header line
    :type h2: str
    """
    def __init__(self, h1:str, h2:str):
        super().__init__(name=h1)

    def encode(self, cell_contents):
        return str(cell_contents)

    def is_data(self):
        return True

    def is_hpo(self):
        return False

class HpoEncoder(CellEncoder):
    """Convenience class to represent the header (two lines) of columns that represent HPO columns in the template

    :param h1: contents of the first header line
    :type h1: str
    :param h2: contents of the second header line
    :type h2: str
    """
    def __init__(self, h1:str, h2:str):
        super().__init__(name=h1)
        self._error = None
        self._hpo_label = None
        self._hpo_id = None
        if h1.endswith(" "):
            self._error = f"Error - HPO label ends with whitespace: \”{h1}\""
        elif h2.startswith("HP:") and len(h2) != 10:
            self._error = f"Error - Malformed HPO id: \”{h2}\" ({h1})"
        elif not h2.startswith("HP:"):
            self._error = f"Error - {h1} -- {h2}"
        else:
            self._hpo_label = h1
            self._hpo_id = h2

    def is_data(self):
        return False

    def is_hpo(self):
        """
        :returns: True iff this is an NPO column and there was no error
        :rtype: bool
        """
        return self._error is None

    def needs_attention(self):
        """
        :returns: True iff there was a problem with this column
        :rtype: bool
        """
        return self._error is not None

    def get_error(self):
        return self._error

    def encode(self, cell_contents):
        cell_contents = str(cell_contents)
        if cell_contents == "observed":
            return HpTerm(hpo_id=self._hpo_id, label=self._hpo_label)
        elif cell_contents == "excluded":
            return  HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, observed=False)
        elif cell_contents.startswith("P"):
            # iso8601 age
            return  HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, onset=cell_contents)
        elif cell_contents == "na" or cell_contents == "nan" or len(cell_contents) == 0:
            return None
        else:
            raise ValueError(f"Could not parse HPO column cell_contents: \”{cell_contents}\"")

class NullEncoder(CellEncoder):
    """Convenience class to represent a column that we do not use for encoding

    :param h1: contents of the first header line
    :type h1: str
    :param h2: contents of the second header line
    :type h2: str
    """
    def __init__(self, h1=None, h2=None):
        super().__init__(name="Begin of HPO column")

    def encode(self, cell_contents):
        return None

    def is_data(self):
        return False

    def is_hpo(self):
        return False

EXPECTED_HEADERS = {"PMID", "title", "individual_id", "Comment", "disease_id", "disease_label", "transcript",
                            "allele_1", "allele_2", "variant.comment", "age_of_onset", "sex"}

DATA_ITEMS = {"PMID", "title", "individual_id", "disease_id", "disease_label", "transcript",
                            "allele_1", "allele_2",  "age_of_onset", "sex"}




class CaseTemplateEncoder:
    """Class to encode data from user-provided Excel template.

    :param df: template table with clinical data
    :type df: pd.DataFrame
    :param hpo_cr: HpoConceptRecognizer for text mining
    :type hpo_cr: pyphetools.creation.HpoConceptRecognizer
    :param created_by: biocurator (typically, this should be an ORCID identifier)
    :type created_by: str
    """

    HPO_VERSION = None

    def __init__(self, df:pd.DataFrame, hpo_cr:HpoConceptRecognizer, created_by:str) -> None:
        """constructor
        """
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"argument \"df\" must be pandas DataFrame but was {type(df)}")
        self._individuals = []
        self._errors = []
        header_1 = df.columns.values.tolist()
        header_2 = df.loc[0, :].values.tolist()
        if len(header_1) != len(header_2):
            # should never happen unless the template file is corrupted
            raise ValueError("headers are different lengths. Check template file for correctness.")
        self._n_columns = len(header_1)
        self._index_to_decoder = self._process_header(header_1, header_2)
        data_df = df.iloc[1:]
        self._is_biallelic = "allele_2" in header_1
        self._allele1_d = {}
        self._allele2_d = {}
        for _, row in data_df.iterrows():
            individual = self._parse_individual(row)
            self._individuals.append(individual)
            self._allele1_d[individual.id] = row["allele_1"]
            if self._is_biallelic:
                self._allele2_d[individual.id] = row["allele_2"]
        CaseTemplateEncoder.HPO_VERSION = hpo_cr.get_hpo_ontology().version
        self._created_by = created_by
        self._metadata_d = {}
        for i in self._individuals:
            cite = i._citation
            metadata = MetaData(created_by=created_by, citation=cite)
            metadata.default_versions_with_hpo(CaseTemplateEncoder.HPO_VERSION)
            self._metadata_d[i.id] = metadata

    def  _process_header(self, header_1:List, header_2:List) -> Dict[int, CellEncoder]:
        index_to_decoder_d = {}
        in_hpo_range = False
        for i in range(self._n_columns):
            h1 = header_1[i]
            h2 = header_2[i]
            if h1 == "HPO":
                in_hpo_range = True
                index_to_decoder_d[i] = NullEncoder()
                continue
            elif not in_hpo_range and h1 in EXPECTED_HEADERS:
                index_to_decoder_d[i] = DataEncoder(h1=h1, h2=h2)
            elif in_hpo_range:
                encoder = HpoEncoder(h1=h1, h2=h2)
                if encoder.needs_attention():
                    self._errors.append(encoder.get_error())
                    index_to_decoder_d[i] = NullEncoder()
                else:
                    index_to_decoder_d[i] = encoder
        if not in_hpo_range:
            raise ValueError("Did not find HPO boundary column")
        print(f"Created encoders for {len(index_to_decoder_d)} fields")
        return index_to_decoder_d

    def _parse_individual(self, row:pd.Series):
        if not isinstance(row, pd.Series):
            raise ValueError(f"argument df must be pandas DSeriestaFrame but was {type(row)}")
        data = row.values.tolist()
        if len(data) != self._n_columns:
            # Should never happen
            raise ValueError(f"Divergent number of columns: header {self._n_columns} but data row {len(data)}: {data}")
        data_items = {}
        hpo_terms = list()
        for i in range(self._n_columns):
            encoder = self._index_to_decoder.get(i)
            cell_contents = data[i]
            if encoder is None:
                print(f"Encoder {i} was None for data {cell_contents}")
                continue
            if encoder.is_data() and encoder.name in DATA_ITEMS:
                data_items[encoder.name] = encoder.encode(cell_contents)
            elif encoder.is_hpo():
                hpoterm = encoder.encode(cell_contents)
                if hpoterm is not None:
                    hpo_terms.append(hpoterm)
        # Check we have all of the items we need
        for item in data_items.keys():
            if item not in DATA_ITEMS:
                raise ValueError(f"Unrecognized data item: \"{item}\"")
        #Note that allele_2 is optional
        if len(data_items) < len(DATA_ITEMS) - 1:
            raise ValueError(f"Insufficient data items: \"{data_items}\"")
        # If we get here, we can contruct an individual
        individual_id = data_items.get('individual_id')
        pmid = data_items.get("PMID")
        title = data_items.get("title")
        citation = Citation(pmid=pmid, title=title)
        sex = data_items.get("sex")
        if sex == "M":
            sex = Constants.MALE_SYMBOL
        elif sex == "F":
            sex = Constants.FEMALE_SYMBOL
        elif sex == "O":
            sex = Constants.OTHER_SEX_SYMBOL
        elif sex == "U":
            sex = Constants.UNKOWN_SEX_SYMBOL
        else:
            raise ValueError(f"Unrecognized sex symbol: {sex}")
        age = data_items.get(AGE_OF_ONSET_FIELDNAME)
        if age is not None and isinstance(age, str) and age.startswith("P"):
            isoage = age
        else:
            isoage = Constants.NOT_PROVIDED
        disease_id = data_items.get("disease_id")
        disease_label = data_items.get("disease_label")
        disease = Disease(disease_id=disease_id, disease_label=disease_label)
        return Individual(individual_id=individual_id,
                            citation=citation,
                            hpo_terms=hpo_terms,
                            sex=sex,
                            age=isoage,
                            disease=disease)

    def get_individuals(self) -> List[Individual]:
        return self._individuals

    def get_allele1_d(self)-> Dict[str,str]:
        return self._allele1_d

    def get_allele2_d(self)-> Dict[str,str]:
        return self._allele2_d

    def _is_biallelic(self) -> bool:
        return self._is_biallelic

    def get_metadata_d(self) -> Dict[str,MetaData]:
        return self._metadata_d

    def get_phenopackets(self) -> List[PPKt.Phenopacket]:
        ppack_list = []
        for individual in self._individuals:
            cite = individual._citation
            metadata = MetaData(created_by=self._created_by, citation=cite)
            metadata.default_versions_with_hpo(CaseTemplateEncoder.HPO_VERSION)
            phenopckt = individual.to_ga4gh_phenopacket(metadata=metadata)
            ppack_list.append(phenopckt)
        return ppack_list



    def output_individuals_as_phenopackets(self, individual_list:List[Individual], outdir="phenopackets") -> None:
        """write a list of Individual objects to file in GA4GH Phenopacket format
        Note that the individual_list needs to be passed to this object, because we expect that
        the QC code will have been used to cleanse the data of redundancies etc before output.
        We use the statefullness to keep track of the created_by argument from the constructor

        :param outdir: Path to output directory. Defaults to "phenopackets". Created if not exists.
        :type outdir: str
        """
        if os.path.isfile(outdir):
            raise ValueError(f"Attempt to create directory with name of existing file {outdir}")
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        written = 0

        if self._created_by is None:
            created_by = 'pyphetools'
        else:
            created_by = self._created_by
        for individual in individual_list:
            cite = individual._citation
            metadata = MetaData(created_by=created_by, citation=cite)
            metadata.default_versions_with_hpo(CaseTemplateEncoder.HPO_VERSION)
            phenopckt = individual.to_ga4gh_phenopacket(metadata=metadata)
            json_string = MessageToJson(phenopckt)
            pmid = cite.pmid
            if pmid is None:
                fname = "phenopacket_" + individual.id
            else:
                pmid = pmid.replace(" ", "").replace(":", "_")
                fname = pmid + "_" + individual.id
            fname = re.sub('[^A-Za-z0-9_-]', '', fname)  # remove any illegal characters from filename
            fname = fname.replace(" ", "_") + ".json"
            outpth = os.path.join(outdir, fname)
            with open(outpth, "wt") as fh:
                fh.write(json_string)
                written += 1
        print(f"We output {written} GA4GH phenopackets to the directory {outdir}")


    def to_summary(self) -> pd.DataFrame:
        """

        The table provides a summary of the table that was parsed from the input file. If there were errors, it
        provides enough feedback so that the user knows what needs to be fixed

        :returns: an table with status of parse
        :rtype: pd.DataFrame
        """
        n_error = 0
        items = []
        for e in self._errors:
            n_error += 1
            d = {'item': f"Error {n_error}", 'value': e}
            items.append(d)
        d = {'item': 'created by', 'value':self._created_by}
        items.append(d)
        d = {'item':'number of individuals', 'value': str(len(self._individuals))}
        items.append(d)
        n_hpo_columns = sum([1 for encoder in self._index_to_decoder.values() if encoder.is_hpo()])
        d = {'item':'number of HPO columns', 'value': str(n_hpo_columns)}
        items.append(d)
        return pd.DataFrame(items)






import abc


class PhenopacketValidator(metaclass=abc.ABCMeta):
    """
    Abstract super class for classes that validate phenopackets
    """
    def __init__(self):
        pass

    @abc.abstractmethod
    def validate_phenopacket(self, phenopacket):
        pass


import unittest

import pytest

from src.pyphetools.creation import MetaData, Citation


class TestMetaData(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        hpo_version = "fake.version"
        pmid = "PMID:30945334"
        title = "Recurrent de novo MAPK8IP3 variants cause neurological phenotypes"
        citation = Citation(pmid=pmid, title=title)
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199", citation=citation)
        metadata.default_versions_with_hpo(version=hpo_version)
        cls._metadata = metadata.to_ga4gh()


    def test_created_by(self):
        expected_orcid = "ORCID:0000-0002-0736-9199"
        self.assertEqual(expected_orcid, self._metadata.created_by)

    def test_hpo_version(self):
        """We should retrieve Intellectual disability, mild (HP:0001256)"""
        mdata = self._metadata
        res = mdata.resources
        hpo_res = None
        for r in res:
            if r.id == "hp":
                hpo_res = r
        self.assertIsNotNone(hpo_res)
        self.assertEqual("fake.version", hpo_res.version)

    def test_has_five_resources(self):
        """
        the method default_versions_with_hpo adds a total of 5 ontologies to the MetaData
        """
        mdata = self._metadata
        self.assertEqual(5, len(mdata.resources))

    def test_external_reference(self):
        mdata = self._metadata
        self.assertEqual(1, len(mdata.external_references))
        eref = mdata.external_references[0]
        self.assertEqual("PMID:30945334", eref.id)
        self.assertEqual("https://pubmed.ncbi.nlm.nih.gov/30945334", eref.reference)
        self.assertEqual("Recurrent de novo MAPK8IP3 variants cause neurological phenotypes", eref.description)

    def test_pmid(self):
        # Test whether the pyphetools MetaData object returns the right PMID
        hpo_version = "fake.version"
        pmid = "PMID:30945334"
        title = "Recurrent de novo MAPK8IP3 variants cause neurological phenotypes"
        citation = Citation(pmid=pmid, title=title)
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199", citation=citation)
        self.assertEqual("PMID:30945334", metadata.get_pmid())


    def test_pmid_not_present(self):
        # Test whether the pyphetools MetaData object throws an error if we try to get a PMID that is not there
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199")
        with pytest.raises(ValueError) as v_e:
            pmid = metadata.get_pmid()

    def test_ga4gh_metadata_pmid(self):
        mdata = self._metadata
        self.assertEqual(1, len(mdata.external_references))
        eref = mdata.external_references[0]
        self.assertEqual("PMID:30945334", eref.id)





from collections import defaultdict
import phenopackets as PPKt
from typing import Dict, List, Set
import sys

from ..creation.constants import Constants
from ..creation import Individual, HpTerm, MetaData
from .simple_patient import SimplePatient
from .html_table_generator import HtmlTableGenerator


class Age2Day:
    """
    Convenience function to help sort ages but converting ISO8601 strings into the number of days

        sorted_age = sorted(l, key=lambda x: x.days)

    following this, retrieve the original string as x.key
    """
    def __init__(self, age, days) -> None:
        self.key = age
        self.days = days


class IndividualTable:

    """
    This class creates a table with a summary of  data in a cohort of individuals
    Create Individual objects and transform them into phenopackets, or import GA4GH phenopackets and display them.

        from IPython.display import HTML, display
        phenopackets = [i.to_ga4gh_phenopacket(metadata=metadata) for i in individuals]
        table = PhenopacketTable.from_phenopackets(phenopacket_list=phenopackets)
        display(HTML(table.to_html()))

    Alternatively,

        table = PhenopacketTable.from_individuals(individual_list=individuals, metadata=metadata)
        display(HTML(table.to_html()))
    """
    def __init__(self,  individual_list:List[Individual],
                        metadata:MetaData=None) -> None:
        """
        :param phenopacket_list: List of GA4GH phenopackets to be displayed
        :type phenopacket_list: List[PPKt.Phenopacket]
        :param individual_list: List of Indidivual objects to be displayed
        :type individual_list: List[Individual]
        :param metadata: metadata about the individuals
        :type metadata: MetaData
        """
        self._spat_list = []
        pmid_count_d = defaultdict(int)
        no_pmid_found = 0
        pmid_found = 0
        for individual in individual_list:
            pp = self._individual_to_phenopacket(individual, metadata)
            spat = SimplePatient(ga4gh_phenopacket=pp)
            if spat.has_pmid():
                pmid_count_d[spat.get_pmid()] += 1
                pmid_found += 1
            else:
                no_pmid_found += 1
            self._spat_list.append(spat)
        # Create caption
        if pmid_found == 0:
            self._caption = f"{len(individual_list)} individuals with no PMIDs (consider adding this information to the MetaData)"
        else:
            pmid_strings = []
            for k, v in pmid_count_d.items():
                pmid_strings.append(f"{k} (n={v})")
            pmid_str = "; ".join(pmid_strings)
            n_phenopackets = len(individual_list)
            if n_phenopackets == 1:
                self._caption = f"{n_phenopackets} phenopacket - {pmid_str}"
            else:
                self._caption = f"{n_phenopackets} phenopackets - {pmid_str}"


    def to_html(self):
        header_items = ["Individual", "Disease", "Genotype", "Phenotypic features"]
        rows = []
        for spat in self._spat_list:
            rows.append(self._simple_patient_to_table_row(spat))
        generator = HtmlTableGenerator(caption=self._caption, header_items=header_items, rows=rows)
        return generator.get_html()


    def _individual_to_phenopacket(self, individual, metadata):
        """Create a phenopacket with the information from this individual

        We try to get information about the publication from the Individual object first. If this is not
        available, we look for citation information in the metadata

        Args:
            individual (_type_): _description_
            metadata (_type_): _description_

        Raises:
            ValueError: _description_
        """
        citation = individual.get_citation()
        if citation is None:
            if  metadata is None:
                raise ValueError("No citation information available in individual and metadata is None")
            else:
                citation = metadata.get_citation()
        created_by = "pyphetools" # Note that this is only used to create a table showing variants, diagnoses, and HPOs
        # the actual created by is not needed for this, and it is user to provide this constant here rather than demand the user enter it.
        # similarly, we do not need to have the version of the HPO here.
        # Note that this does not affect the code that outputs phenopackets, which need to have the correct version of this
        mdata = MetaData(created_by=created_by)
        mdata.default_versions_with_hpo(version="n/a")
        metadata = mdata.to_ga4gh()
        #  the external reference is use to show counts of PMIDs
        extref = PPKt.ExternalReference()
        extref.id = citation.pmid
        pm = citation.pmid.replace("PMID:", "")
        extref.reference = f"https://pubmed.ncbi.nlm.nih.gov/{pm}"
        extref.description = citation.title
        metadata.external_references.append(extref)
        return individual.to_ga4gh_phenopacket(metadata=metadata)

    def _simple_patient_to_table_row(self, spat:SimplePatient) -> List[str]:
        """
        private method intended to create one table row that represents one individual
        :param spat: An object that represents one individual
        :type spat: SimplePatient
        """
        row_items = []
        # Patient information
        pat_info = spat.get_subject_id() + " (" + spat.get_sex() + "; " + spat.get_age() + ")"
        row_items.append( pat_info)
        row_items.append( spat.get_disease())
        # Variant information
        var_list = spat.get_variant_list()
        if len(var_list) == 0:
            row_items.append("n/a" )
        elif len(var_list) == 1:
            var = var_list[0]
            row_items.append( var.get_display() )
        else:
            cell_items = []
            cell_items.append("<ul>")
            for var in var_list:
                cell_items.append("<li>" + var.get_display() + "</li>")
            cell_items.append("</ul>")
            row_items.append( " ".join(cell_items) )
        # HPO information
        hpo_html = self.get_hpo_cell(spat.get_term_by_age_dict())
        row_items.append( hpo_html )
        return row_items


    def get_hpo_cell(self, term_by_age_dict:Dict[str,HpTerm]) -> str:
        """
        :param term_by_age_dict: A dictionary with key - ISO8601 string, value - list of HpTerm objects
        :type term_by_age_dict: Dict[str,HpTerm]
        :returns: HTML code for the HTML cell
        :rtype: str
        """
        lines = []
        age2day_list = IndividualTable.get_sorted_age2data_list(term_by_age_dict.keys())
        sorted_age = sorted(age2day_list, key=lambda x: x.days)
        for onset in sorted_age:
            hpo_list = term_by_age_dict.get(onset.key)
            hpos = "; ".join([hpo.__str__() for hpo in hpo_list])
            if onset.key == Constants.NOT_PROVIDED:
                lines.append(hpos)
            else:
                lines.append(f"<b>{onset.key}</b>: {hpos}")
        return "<br/>".join(lines)

    @staticmethod
    def iso_to_days(iso_age:str) -> int:
        """
        Transform the ISO8601 age strings (e.g., P3Y2M) into the corresponding number of days to facilitate sorting.

        Note that if age is not provided we want to sort it to the end of the list so we transform to a very high number of days.

        :param iso_age: ISO8601 age string (e.g., P3Y2M)
        :type iso_age: str
        :returns: number of days
        :rtype: int
        """
        if iso_age == Constants.NOT_PROVIDED:
            days = sys.maxsize
        elif not iso_age.startswith("P"):
            raise ValueError(f"Invlaid age string: {age}")
        else:
            days = 0
            age = iso_age[1:]
            N = len(age)
            y = age.find("Y")
            if y != -1:
                days = days + int(365.25*int(age[:y]))
                age = age[y+1:]
            m = age.find("M")
            if m != -1:
                days = days + int(30.436875*int(age[:m]))
                age = age[m+1:]
            d = age.find("D")
            if d != -1:
                days = days + int(age[:d])
        return days

    @staticmethod
    def get_sorted_age2data_list(ages:Set[str]) -> List[Age2Day]:
        """
        Create a sorted list of Age2Day objects that we use to display the age in the HTML output.

        :param ages: A set of ISO 8601 age strings
        :type ages: Set[str]
        :returns: A list of sorted Age2Day objects
        :rtype:  List[Age2Day]
        """
        age2day_list = list(Age2Day(age, IndividualTable.iso_to_days(age)) for age in ages)
        sorted_list = sorted(age2day_list, key=lambda x: x.days)
        return sorted_list



from typing import List
from collections import defaultdict
import hpotk
from ..creation.variant import Variant

from .html_table_generator import HtmlTableGenerator




class VariantVisualizer:
    """Helper class to provide additional context and order for displaying summaries of Qc events

    :param level: ERROR, WARNING, or INFORMATION
    :type level: str
    :param category: from the ValidationResult categories
    :type category: str
    """
    def __init__(self, variant_list:List[Variant]) -> None:
        self._variant_list = variant_list


    def to_html(self) -> str:
        validated_individual_list = self._cohort_validator.get_validated_individual_list()
        html_lines = []
        n_individuals = len(validated_individual_list)
        n_individuals_with_errors = sum([1 for i in validated_individual_list if i.has_error()])
        html_lines.append("<h2>Cohort validation</h2>")
        if n_individuals_with_errors == 0:
            html_lines.append(f"<p>No errors found for the cohort with {n_individuals} individuals</p>")
        else:
            para = f"<p>Errors found with {n_individuals_with_errors} of {n_individuals} phenopackets.</p>"
            html_lines.append(para)
            errors = []
            for vi in validated_individual_list:
                if vi.has_error():
                    errors.extend(vi.get_validation_errors())
            errors = sorted(errors, key=lambda x : (x._error_level, x._category, x._message))
            header_fields = ValidationResult.get_header_fields()
            rows = [row.get_items_as_array() for row in errors]
            generator = HtmlTableGenerator(header_items=header_fields, rows=rows, caption="Error analysis")
            html_lines.append(generator.get_html())
        return "\n".join(html_lines)


from hpotk.model import TermId


class TermCount:

    def __init__(self, tid) -> None:
        """Counts the number of usages of an ontology term

        Args:
            tid: An ontology term identifier
        """
        if not isinstance(tid, TermId):
            raise ValueError(f"id argument must be TermId but was {type(tid)}")
        self._hpo_id = tid
        self._count = 0

    def increment(self):
        self._count += 1

    @property
    def hpo_id(self):
        return self._hpo_id

    @property
    def count(self):
        return self._count



from collections import defaultdict
from .phenopacket_ingestor import PhenopacketIngestor
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np



class PhenopacketCharts:

    def __init__(self, indir) -> None:
        ingestor = PhenopacketIngestor(indir=indir)
        simple_patient_list = list(ingestor.get_simple_patient_dictionary().values())
        self._disease_d = defaultdict(int)
        self._pmid_d = defaultdict(int)
        self._hpo_count_d = defaultdict(int) # total count
        self._term_count_d = defaultdict(int)

        for sp in simple_patient_list:
            self._disease_d[sp.get_disease()] += 1
            if sp.has_pmid():
                self._pmid_d[sp.get_pmid()] += 1
            self._hpo_count_d[sp.get_total_hpo_count()] += 1
            for hp in sp.get_observed_hpo_d().values():
                self._term_count_d[hp.label] += 1

    def disease_barchart(self):
        disease = []
        count = []
        for k,v in self._disease_d.items():
            disease.append(k)
            count.append(v)
        df = pd.DataFrame ({
            'Disease': disease,
            'Count': count
        })

        df = df.sort_values(by=['Count'])

        # Create horizontal bars
        ax = plt.barh(y=df.Disease, width=df.Count, height=0.2);

        # Add title
        plt.title('Diseases');
        return ax


    def pmid_barchart(self):
        pmid = []
        count = []
        for k,v in self._pmid_d.items():
            pmid.append(k)
            count.append(v)
        df = pd.DataFrame ({
            'PMID': pmid,
            'Count': count
        })

        df = df.sort_values(by=['Count'])

        # Create horizontal bars
        ax = plt.barh(y=df.PMID, width=df.Count, height=0.2);

        # Add title
        plt.title('Individuals per Citation');
        return ax


    def terms_per_phenopacket(self):
        """barchart with y (height)=number of terms per phenopacket and x: number of phenopackets with that many terms
        """
        max_terms = max(self._hpo_count_d.keys())

        term_vector = []
        n_packet_vector = []
        for i in range(max_terms+1):
            n_packets = self._hpo_count_d.get(i, 0)
            term_vector.append(i)
            n_packet_vector.append(n_packets)
        ax = plt.bar(term_vector, n_packet_vector,color ='green', alpha = 0.7)
        # the following two lines cause the Y axis to show just integers for 20 or less
        max_n_individual = max(self._hpo_count_d.values())
        if max_n_individual < 21:
            yint = range(0, max_n_individual+1)
            plt.yticks(yint)
        plt.xlabel('HPO terms per individual')
        plt.ylabel('# Individuals')
        plt.title("Total HPO term counts per individual")
        return ax

    def most_common_hpo_terms(self, max_terms_to_show=10):
        sorted_terms = sorted(self._term_count_d.items(), key=lambda x:x[1], reverse=True)
        if len(sorted_terms) > max_terms_to_show:
            sorted_terms = sorted_terms[:max_terms_to_show]
        HPO = []
        count = []
        for st in sorted_terms:
            HPO.append(st[0])
            count.append(st[1])
        df = pd.DataFrame ({
            'HPO': HPO,
            'Count': count
        })

        df = df.sort_values(by=['Count'])

        # Create horizontal bars
        ax = plt.barh(y=df.HPO, width=df.Count, height=0.2);

        # Add title
        plt.title('Most common HPO annotations');
        return ax



import abc
import os
from typing import List
from .individual import Individual
from .metadata import MetaData


class AbstractEncoder(metaclass=abc.ABCMeta):
    """
    Abstract superclass for all chort encoder classes

    :param metadata: A pyphetools MetaData object (from which a GA4GH MetaData  object can be constructed).
    :type metadata: MetaData
    """

    def __init__(self, metadata:MetaData) -> None:
        if metadata is None:
            raise ValueError("Must pass a metadata object to constructor")
        # the following workaround is needed because isinstance gets confused about the two following classes
        elif str(type(metadata)) == "<class 'pyphetools.creation.metadata.MetaData'>":
            self._metadata = metadata
        else:
            raise ValueError(F"metadata argument must be pyphetools MetaData but was {type(metadata)}")

    def output_phenopackets(self, outdir:str="phenopackets"):
        """Output data about all individuals as GA4GH phenopackets

        :param outdir: name of directory to write phenopackets (default: 'phenopackets')
        :type outdir: str
        """
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        individual_list = self.get_individuals()
        written = Individual.output_individuals_as_phenopackets(individual_list=individual_list,
                                                                metadata=self._metadata.to_ga4gh(),
                                                                outdir=outdir)
        print(f"Wrote {written} phenopackets to {outdir}")

    @abc.abstractmethod
    def get_individuals(self) -> List[Individual]:
        """Get a list of all Individual objects in the cohort

        :returns: a list of all Individual objects in the cohort
        :rtype: List[Individual]
        """
        raise NotImplementedError("Need to implement in subclass")



import phenopackets
from .variant import Variant
import string
from typing import Dict
import random

ACCEPTABLE_GENOMES = {"GRCh37", "GRCh38", "hg19", "hg38"}



# {'alt': 'C', 'chr': '16', 'pos': '1756403', 'ref': 'CG'}},

class HgvsVariant(Variant):
    """
    This encapsulates variant data that we retrieve from Variant Validator

    :param assembly: the genome build (one of hg19, hg38)
    :type assembly: str
    :param vcf_d: dictionary with values for chr, pos, ref, alt (VCF)
    :type vcf_d: Dict[str]
    :param symbol: the gene symbol
    :type symbol: str, optional
    :param hgnc: The Human Gene Nomenclature Comittee (HNGC) identifier, e.g. HGNS:123
    :type hgnc: str, optional
    :param transcript: identifier of the transcript for defininf the variant
    :type transcript: str, optional
    :param g_hgvs: genomic hgvs
    :type g_hgvs: str, optional
    :param variant_id: variant identifier
    :type variant_id: str, optional
    """
    def __init__(self, assembly, vcf_d, symbol=None, hgnc=None, hgvs=None, transcript=None, g_hgvs=None,
                 variant_id=None) -> None:
        super().__init__()
        if not assembly in ACCEPTABLE_GENOMES:
            raise ValueError(f"Malformed assembly: \"{assembly}\"")
        self._assembly = assembly
        if not isinstance(vcf_d, dict):
            raise ValueError(f"vcf_d argument must be dictionary")
        self._chr = vcf_d.get('chr')
        self._position = int(vcf_d.get('pos'))
        self._ref = vcf_d.get('ref')
        self._alt = vcf_d.get('alt')
        self._symbol = symbol
        self._hgnc_id = hgnc
        self._hgvs = hgvs
        self._transcript = transcript
        self._g_hgvs = g_hgvs
        self._genotype = None
        if variant_id is None:
            self._variant_id = "var_" + "".join(random.choices(string.ascii_letters, k=25))
        else:
            self._variant_id = variant_id

    @property
    def assembly(self):
        return self._assembly

    @property
    def chr(self):
        return self._chr

    @property
    def position(self):
        return self._position

    @property
    def ref(self):
        return self._ref

    @property
    def alt(self):
        return self._alt

    @property
    def genotype(self):
        return self._genotype

    def __str__(self):
        return f"{self._hgvs}({self._chr}:{self._position}{self._ref}>{self._alt})"

    def to_string(self):
        return self.__str__()

    def to_ga4gh_variant_interpretation(self, acmg=None):
        """For the new interface. Refactor client code to use this function, which has an unambiguous name
        """
        return self.to_ga4gh(acmg=acmg)

    def to_ga4gh(self, acmg=None):
        """
        Transform this Variant object into a "variantInterpretation" message of the GA4GH Phenopacket schema
        """
        vdescriptor = phenopackets.VariationDescriptor()
        vdescriptor.id = self._variant_id
        if self._hgnc_id is not None and self._symbol is not None:
            vdescriptor.gene_context.value_id = self._hgnc_id
            vdescriptor.gene_context.symbol = self._symbol
        hgvs_expression = phenopackets.Expression()
        if self._hgvs is not None:
            hgvs_expression.syntax = "hgvs.c"
            hgvs_expression.value = self._hgvs
            vdescriptor.expressions.append(hgvs_expression)
        if self._g_hgvs is not None:
            hgvs_expression.syntax = "hgvs.g"
            hgvs_expression.value = self._g_hgvs
            vdescriptor.expressions.append(hgvs_expression)
        vdescriptor.molecule_context =  phenopackets.MoleculeContext.genomic
        if self._genotype is not None:
            if self._genotype == 'heterozygous':
                vdescriptor.allelic_state.id = "GENO:0000135"
                vdescriptor.allelic_state.label = "heterozygous"
            elif self._genotype == 'homozygous':
                vdescriptor.allelic_state.id = "GENO:0000136"
                vdescriptor.allelic_state.label = "homozygous"
            elif self._genotype == 'hemizygous':
                vdescriptor.allelic_state.id = "GENO:0000134"
                vdescriptor.allelic_state.label = "hemizygous"
            else:
                print(f"Did not recognize genotype {self._genotype}")
        vinterpretation = phenopackets.VariantInterpretation()
        if acmg is not None:
            if acmg.lower() == 'benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.BENIGN
            elif acmg.lower == 'likely benign' or acmg.lower() == 'likely_benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_BENIGN
            elif acmg.lower == 'uncertain significance' or acmg.lower() == 'uncertain_significance':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.UNCERTAIN_SIGNIFICANCE
            elif acmg.lower == 'likely pathogenic' or acmg.lower() == 'likely_pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_PATHOGENIC
            elif acmg.lower == 'pathogenic' or acmg.lower() == 'pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.PATHOGENIC
            else:
                print(f"Warning- did not recognize ACMG category {acmg}")
        vcf_record = phenopackets.VcfRecord()
        vcf_record.genome_assembly =  self._assembly
        vcf_record.chrom = self._chr
        vcf_record.pos = self._position
        vcf_record.ref = self._ref
        vcf_record.alt = self._alt
        vdescriptor.vcf_record.CopyFrom(vcf_record)
        vinterpretation.variation_descriptor.CopyFrom(vdescriptor)
        return vinterpretation



import unittest
import os
from src.pyphetools.creation import HpoParser, OptionColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestOptionColumnMapper2(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()

    def test_multiple_items_in_cell_1(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                    'Infantile spasms': "Infantile spasms" ,
                    'GTC': 'Bilateral tonic-clonic seizure' ,
                    'ESES': "Status epilepticus"
                    }
        seizureMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=seizure_d)
        results = seizureMapper.map_cell("Absence")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0011147", result.id)
        self.assertEqual("Typical absence seizure", result.label)
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)

    def test_multiple_items_in_cell_2(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                    'Infantile spasms': "Infantile spasms",
                    'GTC': 'Bilateral tonic-clonic seizure',
                    'ESES': "Status epilepticus"}
        seizureMapper = OptionColumnMapper(column_name="placeholder",
                                        concept_recognizer=self.hpo_cr,
                                        option_d=seizure_d)
        results = seizureMapper.map_cell("Absence and GTC")
        self.assertEqual(2, len(results))
        results = sorted(results, key=lambda x: x.id, reverse=True)
        result = results[0]
        self.assertEqual("HP:0011147", result.id)
        self.assertEqual("Typical absence seizure", result.label)
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)
        result = results[1]
        self.assertEqual("HP:0002069", result.id)
        self.assertEqual("Bilateral tonic-clonic seizure", result.label)
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)

    def test_multiple_items_in_cell_3(self):
        morph_d = {
            'bulbous nasal tip': "Bulbous nose",
            'prominent lobule of ear': "Large earlobe",
            'tapering fingers': "Tapered finger",
            'deeply set eyes': 'Deeply set eye'
        }
        cell_contents = "Broad forehead, deeply set eyes, ptosis, bulbous nasal tip, micrognathia, prominent lobule of ear, tapering fingers"
        morphMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=morph_d)
        results = morphMapper.map_cell(cell_contents)
        self.assertEqual(7, len(results))
        results = sorted(results, key=lambda x: x.id, reverse=True)

    def test_hpo_cr_non_null(self):
        """sanity check"""
        self.assertIsNotNone(self.hpo_cr)

    def test_hpo_cr_ataxia(self):
        """We should retrieve Ataxia (HP:0001251)"""
        neuroMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d={})
        results = neuroMapper.map_cell("ataxia")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001251", result.id)
        self.assertEqual("Ataxia", result.label)

    def test_hpo_cr_spastic_paraplegia(self):
        """We should retrieve Spastic paraplegia (HP:0001258)"""
        neuroMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d={})
        results = neuroMapper.map_cell("spastic paraplegia")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001258", result.id)
        self.assertEqual("Spastic paraplegia", result.label)

    def test_hpo_cr_multiple_concepts_with_custom_map(self):
        text = 'spasticity; nerve conduction and EMG studies with abnormal findings "remarkable for the failure to activate the leg muscles due to an upper motor neuron pattern of aberrant motor unit potential firing rates. These findings are consistent with dysfunction of the corticospinal pathways rather than a lower motor unit."'
        neuroMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d={})
        results = neuroMapper.map_cell(text)
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001257", result.id)
        self.assertEqual("Spasticity", result.label)

    def test_hpo_cr_multiple_concepts_no_custom_map(self):
        text = 'spasticity; nerve conduction and EMG studies with abnormal findings "remarkable for the failure to activate the leg muscles due to an upper motor neuron pattern of aberrant motor unit potential firing rates. These findings are consistent with dysfunction of the corticospinal pathways rather than a lower motor unit." Significant low extremity weakness.'
        neuro_exam_custom_map = {'low extremity weakness': 'Lower limb muscle weakness',
                                'unstable gait': 'Unsteady gait',
                                'dysfunction of the corticospinal pathways': 'Upper motor neuron dysfunction',
                                }
        neuroMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=neuro_exam_custom_map)
        results = neuroMapper.map_cell(text)
        self.assertEqual(3, len(results))
        term_d = dict([(hpo_term.id, hpo_term.label) for hpo_term in results])
        self.assertTrue("HP:0001257" in term_d)
        self.assertEqual("Spasticity", term_d.get("HP:0001257"))
        self.assertTrue("HP:0007340" in term_d)
        self.assertEqual("Lower limb muscle weakness", term_d.get("HP:0007340"))
        self.assertTrue("HP:0002493" in term_d)
        self.assertEqual("Upper motor neuron dysfunction", term_d.get("HP:0002493"))

    def test_SETD2(self):
        text = """Extra fluid in the back of the cerebellum at 35 weeks; fetal MRI at 35 weeks showed VSD,
        small cerebellum, and agenesis of the corpus callosum; pre-eclampsia; \nIUGR 	n/a """
        prenatal_custom_map = {'agenesis of the corpus callosum': 'Agenesis of corpus callosum',
                            '\nIUGR': 'Intrauterine growth retardation',
                            'small cerebellum': 'Cerebellar hypoplasia',
                            }
        prenatalMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=prenatal_custom_map)
        results = prenatalMapper.map_cell(text)
        self.assertEqual(3, len(results))

    def test_get_maximal_match(self):
        text = "severe global developmental delay"
        dev_custom_map = {'Severe global developmental delay': 'Severe global developmental delay'}
        devMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=dev_custom_map)
        results = devMapper.map_cell(text)
        self.assertEqual(1, len(results))
        res = results[0]
        self.assertEqual(res.label, "Severe global developmental delay")


    def test_list(self):
        agressivenes_d = {'yes': 'Aggressive behavior',
                        'auto and heteroagressivity': ['Self-injurious behavior', 'Aggressive behavior']
                        }
        excluded_D = {"no":'Aggressive behavior' }
        mapper = OptionColumnMapper(column_name="placeholder",
                                    concept_recognizer=self.hpo_cr,
                                    option_d=agressivenes_d,
                                    excluded_d=excluded_D)
        results = mapper.map_cell("yes")
        self.assertEqual(1, len(results))
        res = results[0]
        self.assertEqual(res.label, "Aggressive behavior")
        self.assertEqual(res.id, "HP:0000718")
        results = mapper.map_cell("auto and heteroagressivity")
        self.assertEqual(2, len(results))
        results = sorted(results, key=lambda x: x.id, reverse=True)
        # Self-injurious behavior HP:0100716
        res0 = results[0]
        self.assertEqual("HP:0100716", res0.id)
        res1 = results[1]
        self.assertEqual("HP:0000718", res1.id)

    def test_obs_exc(self):
        """
        The text "muscular hypotonia in the first years" was being recorded as both observed and excluded "Hypotonia"
        """
        hypotonia_d = {'yes hyper elastic': 'Hypotonia',
                        'yes': 'Hypotonia',
                        'axial tone more affected than  appendicular': 'Axial hypotonia',
                        'muscular hypotonia in the first years': 'Hypotonia'}
        hypotoniaMapper = OptionColumnMapper(column_name="placeholder",
                                            concept_recognizer=self.hpo_cr, option_d=hypotonia_d,
                                            excluded_d={'no': 'Hypotonia'})
        results = hypotoniaMapper.map_cell("muscular hypotonia in the first years")
        self.assertEqual(1, len(results))

    def test_obs_excluded_ventriculomegaly(self):
        """
        The text "no ventriculomegaly" was being recorded as both observed and excluded "Ventriculomegaly"
        """
        ventricles_d = {
            'moderate enlargement of the lateral ventricles': 'Lateral ventricle dilatation',
            'left greater than right': 'Lateral ventricle dilatation',
        }
        ventriclesMapper = OptionColumnMapper(column_name="placeholder",
                                            concept_recognizer=self.hpo_cr, option_d=ventricles_d,
                                            excluded_d={"normal": "Ventriculomegaly", 'no ventriculomegaly': 'Ventriculomegaly'})
        results = ventriclesMapper.map_cell("no ventriculomegaly")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("Ventriculomegaly", result.label)
        self.assertEqual("HP:0002119", result.id)
        self.assertFalse(result.observed)




import phenopackets as PPKt
import re
import os
from typing import List, Union
from google.protobuf.json_format import MessageToJson
from .citation import Citation
from .constants import Constants
from .disease import Disease
from .hp_term import HpTerm
from .hgvs_variant import Variant
from .metadata import MetaData, Resource


class Individual:
    """
    A class to represent one individual of the cohort

    :param individual_id: The individual identifier
    :type individual_id: str
    :param hpo_terms: List of HpTerm objects
    :type hpo_terms: List[pyphetools.creation.HpTerm]
    :param sex: String corresponding to the sex of the individual, default, n/a
    :type sex: str
    :param age: String corresponding to the age of the individual (ISO), default, n/a
    :type age: str
    :param interpretation_list: list of GA4GH VariationInterpretation objects
    :type interpretation_list: List[PPKt.VariationInterpretation], optional
    :param disease: object defining the disease diagnosos
    :type disease: Disease, optional
    """

    def __init__(self,
                individual_id:str,
                hpo_terms:List[HpTerm]=None,
                citation:Citation=None,
                sex:str=Constants.NOT_PROVIDED,
                age:str=Constants.NOT_PROVIDED,
                interpretation_list:List[PPKt.VariantInterpretation]=None,
                disease:Disease=None):
        """Constructor
        """
        if isinstance(individual_id, int):
            self._individual_id = str(individual_id)
        elif isinstance(individual_id, str):
            self._individual_id = individual_id
        else:
            raise ValueError(f"individual_id argument must be int or string but was {type(individual_id)}")
        if sex is None:
            self._sex = PPKt.Sex.UNKNOWN_SEX
        else:
            self._sex = sex
        self._age = age
        if hpo_terms is None:
            self._hpo_terms = list()
        else:
            self._hpo_terms = hpo_terms
        if interpretation_list is None:
            self._interpretation_list = list()
        else:
            self._interpretation_list = interpretation_list
        self._disease = disease
        self._citation = citation
        self._vital_status = None

    @property
    def id(self):
        """
        :returns: the individual identifier
        :rtype: str
        """
        return self._individual_id

    @property
    def sex(self):
        """
        :returns: one of 'MALE', 'FEMALE', 'OTHER', 'UNKNOWN'
        :rtype: str
        """
        return self._sex

    def set_sex(self, sex):
        self._sex = sex

    @property
    def age(self):
        """
        :returns: an iso8601 representation of age
        :rtype: str
        """
        return self._age

    def set_age(self, iso_age):
        self._age = iso_age

    @property
    def hpo_terms(self):
        """
        :returns: a list of observed and excluded HPO terms
        :rtype: List[pyphetools.creation.HpTerm]
        """
        return self._hpo_terms

    @property
    def interpretation_list(self) -> List[PPKt.VariantInterpretation]:
        """
        :returns: a list of GA4GH Genomic Interpretations
        :rtype: List[PPKt.VariantInterpretation]
        """
        return self._interpretation_list

    def add_variant(self, v:Union[Variant, PPKt.VariantInterpretation], acmg:str=None):
        """
        :param v: A Variant obeserved in this individual
        :type v: Union[Variant, PPKt.VariantInterpretation]
        :param acmg: One of the five ACMG pathogenicity categories
        :type acmg: str
        """
        if isinstance(v, Variant):
            variant = v.to_ga4gh_variant_interpretation(acmg=acmg)
        else:
            variant = v
        if isinstance(variant, PPKt.VariantInterpretation):
            self._interpretation_list.append(variant)
        else:
            raise ValueError(f"variant argument must be pyphetools Variant or GA4GH VariantInterpretation but was {type(variant)}")



    def add_hpo_term(self, term:HpTerm) -> None:
        """
        Adds one HPO term to the current individual.
        :param term: An HPO term (observed or excluded, potentially with Age of observation
        :type term: HpTerm
        """
        if not isinstance(term, HpTerm):
            raise ValueError(f"\"term\" argument must be HpTerm but was {type(term)}")
        self._hpo_terms.append(term)

    def set_disease(self, disease:Disease) -> None:
        """
        This method is typically useful for a cohort with multiple diagnoses; otherwise, the disease can be set by the
        CohortEncoder

        :param disease: the disease diagnosis for this individual
        :type disease: Disease
        """
        self._disease = disease

    def set_hpo_terms(self, cleansed_hpo_terms:List[HpTerm]):
        """
        :param cleansed_hpo_terms: a list of HpTerm objects that has been cleansed by OntologyQC
        :type cleansed_hpo_terms: List[pyphetools.creation.HpTerm]
        """
        self._hpo_terms = cleansed_hpo_terms

    @property
    def pmid(self):
        return self._citation.pmid

    def set_citation(self, citation:Citation):
        """
        :param citation: Object with the title and PubMed identifier for the publication in which this individual was described (e.g. PMID:321..)
        :type citation: Citation
        """
        self._citation = citation

    def set_vital_status(self, vstatus:PPKt.VitalStatus):
        self._vital_status = vstatus

    def get_phenopacket_id(self, phenopacket_id=None) -> str:
        """
        :returns: the Phenopacket identifier for this individual
        :rtype: str
        """
        if phenopacket_id is None:
            indi_id = self._individual_id.replace(" ", "_")
            if self._citation is not None:
                pmid = self._citation.pmid.replace(":", "_")
                ppkt_id = f"{pmid}_{indi_id}"
            else:
                ppkt_id = indi_id
        else:
            ppkt_id = phenopacket_id
        ppkt_id = ppkt_id.replace(" ", "_")
        return ppkt_id

    def get_citation(self):
        return self._citation


    def to_ga4gh_phenopacket(self, metadata, phenopacket_id=None) -> PPKt.Phenopacket:
        """
        Transform the data into GA4GH Phenopacket format
        :returns:  a GA4GH Phenopacket representing this individual
        :rtype: PPKt.Phenopacket
        """
        if isinstance(metadata, MetaData):
            metadata = metadata.to_ga4gh()
        if not isinstance(metadata, PPKt.MetaData):
            raise ValueError(f"metadata argument must be pyphetools.MetaData or GA4GH MetaData but was {type(metadata)}")
        php = PPKt.Phenopacket()
        php.id = self.get_phenopacket_id(phenopacket_id=phenopacket_id)
        php.subject.id = self._individual_id
        if self._sex == Constants.MALE_SYMBOL:
            php.subject.sex = PPKt.Sex.MALE
        elif self._sex == Constants.FEMALE_SYMBOL:
            php.subject.sex = PPKt.Sex.FEMALE
        elif self._sex == Constants.OTHER_SEX_SYMBOL:
            php.subject.sex = PPKt.Sex.OTHER_SEX
        elif self._sex == Constants.UNKOWN_SEX_SYMBOL:
            php.subject.sex = PPKt.Sex.UNKNOWN_SEX
        if self._age is not None and self._age != Constants.NOT_PROVIDED:
            php.subject.time_at_last_encounter.age.iso8601duration = self._age
        if self._vital_status is not None:
            php.subject.vital_status.CopyFrom(self._vital_status)
        if isinstance(self._hpo_terms, list):
            for hp in self._hpo_terms:
                if not hp.measured:
                    continue
                pf = hp.to_ga4gh_phenotypic_feature()
                if pf.onset.age.iso8601duration is None and self._age != Constants.NOT_PROVIDED:
                    pf.onset.age.iso8601duration = self._age
                php.phenotypic_features.append(pf)
        elif isinstance(self._hpo_terms, dict):
            for age_key, hpoterm_list in self._hpo_terms.items():
                for hp in hpoterm_list:
                    if not hp.measured:
                        continue
                    pf = hp.to_phenotypic_feature()
                    # only adjust age of onset if not present
                    if pf.onset.age.iso8601duration is None and age_key.startswith("P"):
                        pf.onset.age.iso8601duration = age_key
                    php.phenotypic_features.append(pf)
        if len(self._interpretation_list) > 0:
            interpretation = PPKt.Interpretation()
            interpretation.id = self._individual_id
            interpretation.progress_status = PPKt.Interpretation.ProgressStatus.SOLVED
            if self._disease is not None:
                interpretation.diagnosis.disease.id = self._disease.id
                interpretation.diagnosis.disease.label = self._disease.label
            for var in self._interpretation_list:
                genomic_interpretation = PPKt.GenomicInterpretation()
                genomic_interpretation.subject_or_biosample_id = self._individual_id
                # by assumption, variants passed to this package are all causative
                genomic_interpretation.interpretation_status = PPKt.GenomicInterpretation.InterpretationStatus.CAUSATIVE
                genomic_interpretation.variant_interpretation.CopyFrom(var)
                interpretation.diagnosis.genomic_interpretations.append(genomic_interpretation)
            php.interpretations.append(interpretation)
        if self._citation is not None:
            # overrides the "general" setting of the external reference for the entire cohort
            metadata.external_references.clear()
            extref = PPKt.ExternalReference()
            extref.id = self._citation.pmid
            pm = self._citation.pmid.replace("PMID:", "")
            extref.reference = f"https://pubmed.ncbi.nlm.nih.gov/{pm}"
            extref.description = self._citation.title
            metadata.external_references.append(extref)
        php.meta_data.CopyFrom(metadata)
        return php

    def __str__(self):
        hpo_list = [t.to_string() for t in self._hpo_terms]
        hpo_str = "\n" + "\n".join(hpo_list)
        return f"{self._individual_id}: {self._age}, {self._sex}: {self._disease} {hpo_str}"

    @staticmethod
    def output_individuals_as_phenopackets(individual_list, metadata:MetaData, outdir="phenopackets"):
        """write a list of Individual objects to file in GA4GH Phenopacket format

        This methods depends on the MetaData object having a PMID and will fail otherwise

        :param individual_list: List of individuals to be written to file as phenopackets
        :type individual_list: List[Individual]
        :param metadata: pyphetools MetaData object
        :type metadata: MetaData
        :param outdir: Path to output directory. Defaults to "phenopackets". Created if not exists.
        :type outdir: str
        """
        if os.path.isfile(outdir):
            raise ValueError(f"Attempt to create directory with name of existing file {outdir}")
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        written = 0
        pmid = metadata.get_pmid()
        for individual in individual_list:
            phenopckt = individual.to_ga4gh_phenopacket(metadata=metadata)
            json_string = MessageToJson(phenopckt)
            if pmid is None:
                fname = "phenopacket_" + individual.id
            else:
                pmid = pmid.replace(" ", "").replace(":", "_")
                fname = pmid + "_" + individual.id
            fname = re.sub('[^A-Za-z0-9_-]', '', fname)  # remove any illegal characters from filename
            fname = fname.replace(" ", "_") + ".json"
            outpth = os.path.join(outdir, fname)
            with open(outpth, "wt") as fh:
                fh.write(json_string)
                written += 1
        print(f"We output {written} GA4GH phenopackets to the directory {outdir}")



    @staticmethod
    def from_ga4gh_metadata(mdata:PPKt.MetaData) -> MetaData:
        created_by = mdata.created_by
        created_time = str(mdata.created)
        if len (mdata.external_references) > 1:
            raise ValueError("multiple external references not supported")
        elif len(mdata.external_references) == 0:
            id = None
            reference = None
            description = None
        else:
            eref = mdata.external_references[0]
            id = eref.id
            reference = eref.reference
            description = eref.description
        resource_list = []
        for resource in mdata.resources:
            resource_id=resource.id
            name = resource.name
            namespace_prefix = resource.namespace_prefix
            iri_prefix = resource.iri_prefix
            url = resource.url
            version = resource.version
            r = Resource(resource_id=resource_id,name=name, namespace_prefix=namespace_prefix, iriprefix=iri_prefix, url=url, version=version)
            resource_list.append(r)
        cite = Citation(pmid=id, title=description)
        metadata = MetaData(created_by=created_by, citation=cite)
        for r in resource_list:
            metadata.add_reference(r)
        return metadata

    @staticmethod
    def get_variants_and_disease(ppkt:PPKt.Phenopacket):
        """extract the pyphetools Disease object and the VariantInterpretation objects that can be used to construct an Individual

        :param ppkt: a GA4GH phenopacket
        :type ppkt: PPKT.Phenopacket
        :returns: tjhe corresponding Individual object
        :rtype: Individual, List[PPKt.VariantInterpretation]
        """
        if len(ppkt.interpretations) == 0:
            print(f"No interpretation found for {ppkt.id}")
            return None, []
        if len(ppkt.interpretations) > 1:
            raise ValueError(f"pyphetools dpoes not currently support multiple Interpretation messages in one phenopacket but we found {len(ppkt.interpretations)}")
        interpretation = ppkt.interpretations[0]
        if interpretation.diagnosis is not None and interpretation.diagnosis.disease is not None:
            d = interpretation.diagnosis.disease
            disease = Disease(disease_id=d.id, disease_label=d.label)
        else:
            disease = None
        if len(interpretation.diagnosis.genomic_interpretations) == 0:
            return disease, []
        else :
            variant_list = []
            for gen_interpretation in interpretation.diagnosis.genomic_interpretations:
                variant_list.append(gen_interpretation.variant_interpretation)
            return disease, variant_list


    @staticmethod
    def from_ga4gh_phenopacket(ppkt:PPKt.Phenopacket):
        """
        Transform a GA4GH Phenopacket into an Individual obect -- useful for testing
        :returns:  an individual object corresponding to the GA4GH Phenopacket
        :rtype: Individual
        """
        if not isinstance(ppkt, PPKt.Phenopacket):
            raise ValueError(f"argument must be a GA4GH Phenopacket Message but was {type(ppkt)}")
        #metadata = ppkt.meta_data
        #pypt_metadata = Individual.from_ga4gh_metadata(mdata=metadata)
        subject_id =  ppkt.subject.id
        sex = ppkt.subject.sex
        age = ppkt.subject.time_at_last_encounter.age.iso8601duration
        variant_interpretations = []
        hpo_terms = []
        for pf in ppkt.phenotypic_features:
            hpo_id = pf.type.id
            hpo_label = pf.type.label
            observed = not pf.excluded
            if pf.onset.age.iso8601duration is not None and pf.onset.age.iso8601duration.startswith("P"):
                onset_age = pf.onset.age.iso8601duration
            else:
                onset_age = None
            hpo_terms.append(HpTerm(hpo_id=hpo_id, label=hpo_label, observed=observed, onset=onset_age))
        disease, var_list = Individual.get_variants_and_disease(ppkt)
        indi = Individual(individual_id=subject_id,
                            hpo_terms=hpo_terms,
                            citation=None,
                            sex=sex,
                            age=age,
                            interpretation_list=var_list)
        if disease is not None:
            indi.set_disease(disease=disease)
        return indi




## constants.py
## used to define constants that are used in multiple places in the package
## these constants should not be exported



class Constants:

    MALE_SYMBOL = 'MALE'
    FEMALE_SYMBOL = 'FEMALE'
    OTHER_SEX_SYMBOL = 'OTHER'
    UNKOWN_SEX_SYMBOL = 'UNKNOWN'
    NOT_PROVIDED = 'NOT_PROVIDED'

